<?php

/**
 * Entity controller implementing EntityAPIControllerInterface.
 *
 * Inspired in part by http://wunderkraut.com/blog/remote-entities-in-drupal-7/2012-10-25.
 *
 * This allows loading and saving from the remote source in addition to using
 * the local cached copy.
 *
 * The standard methods try to do the sane thing you'd want to happen in usual
 * circumstances; for other cases, other methods exist.
 *  - load() tries to work locally, and will resort to a remote retrieval if
 *    the entity is expired.
 *  - save() works only locally. This is so that entity forms can work as
 *    normal, and strictly local operations such as saving FieldAPI fields
 *    don't incur a costly remote operation.
 *  - remote_save() saves the entity remotely.
 */
class RemoteEntityAPIDefaultController extends EntityAPIController implements EntityAPIControllerInterface {

  /**
   * Determines whether to bypass remote retrieval of entities.
   *
   * This is needed because several pathways that aren't load() nonetheless
   * result in a load, including saving and deleting. Since we also want to save
   * locally when we retrieve remotely, and we want to purge entities locally
   * without incurring a retrieval, we need this to bypass loading during
   * those operations (the former is circular, the second just slow). Because
   * the results of entity_get_controller() are statically cached, this can be
   * set before saving and deleting and then restored afterwards.
   */
  public $bypass_remote_retrieve = FALSE;

  /**
   * Overridden to retrieve entities from remote source if expired.
   *
   * Retrieving remote entities can be bypassed by setting the value of the
   * property $bypass_remote_retrieve on the controller object to TRUE.
   *
   * @see DrupalDefaultEntityController#load($ids, $conditions)
   */
  public function load($ids = array(), $conditions = array()) {
    $entities = parent::load($ids, $conditions);
    //dsm($entities, 'CONTROLLER load');

    // Check for expiry, allowing the controller setting to bypass this.
    if (!$this->bypass_remote_retrieve) {
      foreach ($entities as $id => $entity) {
        // A value of 0 for the expiry means it doesn't expire.
        if (!empty($entity->expires) && $entity->expires < REQUEST_TIME) {
          // Entity has expired. Fetch the remote copy and update.
          $remote_id = $entity->remote_id;

          $resource = clients_resource_get_for_component('remote_entity', $this->entityType);
          // Load the remote entity. This in turn:
          // - calls the connection to build a RemoteEntityQuery
          // - calls us to take care of doing the packing and the saving.
          $remote_entity = $resource->remote_entity_load($remote_id);
          //dsm($remote_entity, 'controller load fetched entity from remote!');

          // Replace the entity in the array, it is now stale.
          $entities[$id] = $remote_entity;
        }
      }
    }

    return $entities;
  }

  /**
   * Load a local copy of an entity by remote id.
   *
   * This bypasses everything else and should be used to get a local copy when
   * comparing it or associating it to a retrieved remote.
   *
   * @param $remote_id
   *  The remote ID of the entity. This is saved in the 'remote_id' column in
   *  the table, and in the remote entity data is in the key given by the
   *  entity info.
   *
   * @return
   *  Either an entity object (without FieldAPI fields) or NULL if none found.
   */
  public function local_load_by_remote_id($remote_id) {
    $result = $this->query(FALSE, array('remote_id' => $remote_id));
    foreach ($result as $record) {
      // The remote ID is unique; there should be only one.
      $entity = $record;
      $entity->entity_data = unserialize($entity->entity_data);
      return $entity;
    }
  }

  /**
   * Sets the expiry time on the entity object.
   *
   * Note this does not save: call save() afterwards.
   *
   * This is safe to call even if the entity info doesn't use expiry, in which
   * case, calling this will do nothing.
   *
   * @param $entity
   *  The entity to set expiry time on.
   * @param $expiry_time
   *  (optional) The expiry time to set. If not given, this is calculated based
   *  on the request time and the entity info properties.
   *  @see remote_entity_hook_entity_info()
   */
  public function set_expiry($entity, $expiry_time = NULL) {
    // Only set this if the entity info requests it.
    if (isset($this->entityInfo['expiry'])) {
      if (!isset($expiry_time)) {
        // Set the expiry timestamp.
        $expiry_time = REQUEST_TIME + $this->entityInfo['expiry']['expiry time'];
      }
      $entity->expires = $expiry_time;
    }
  }

  /**
   * Save the entity locally.
   *
   * The parent method works fine here. For now!
   */
  //public function save($entity, DatabaseTransaction $transaction = NULL) {}

  /**
   * Process a remote entity that has been retrieved by a RemoteEntityQuery.
   *
   * @param $remote_entities
   *  An array of raw remote entities, as retrieved from the remote connection
   *  with a RemoteEntityQuery.
   *
   * @return
   *  An array of local entities, keyed by entity id.
   *
   * @see clients_resource_remote_entity::executeRemoteEntityQuery()
   */
  public function process_remote_entities($remote_entities) {
    // Set ourselves to bypass remote retrievals, as we don't want the
    // save operation (which calls load()!) to result in circularity.
    $this->bypass_remote_retrieve = TRUE;

    $entites = array();
    foreach ($remote_entities as $remote_entity) {
      // Pack the remote entity into a local entity.
      $entity = $this->pack($remote_entity);

      // Set the expiry time before we save it.
      $this->set_expiry($entity);
      $this->save($entity);

      $entities[$entity->eid] = $entity;
    }

    // Remove the bypass.
    $this->bypass_remote_retrieve = FALSE;

    return $entities;
  }

  /**
   * Packs remote entity data into a local entity for local consumption.
   *
   * (Are you local? ;)
   *
   * This ensures we don't get any clashes between remote and local keys, and
   * also means the entity object we work with locally looks the same as what
   * we stored in the database here.
   *
   * If the entity is already saved locally, it gets associated with the local
   * copy, thus preserving IDs.
   *
   * @param $remote_entity
   *  A remote entity object.
   *
   * @return
   *  A local entity, which has not yet been saved.
   */
  public function pack($remote_entity) {
    $remote_id_key = $this->entityInfo['remote entity keys']['remote id'];
    $remote_id = $remote_entity->{$remote_id_key};

    // Is this entity already stored locally?
    $local_entity = $this->local_load_by_remote_id($remote_id);
    if ($local_entity) {
      // If stored locally, associate the incoming entity with the existing
      // cached copy.
      $local_entity->entity_data = $remote_entity;
      // Set the updated time.
      // TODO: does save() do this anyway???
      $local_entity->updated = REQUEST_TIME;

      $entity = $local_entity;
    }
    else {
      // Create a new stub entity to hold the remote entity.
      $values = array(
        'created' => REQUEST_TIME,
        'updated' => REQUEST_TIME,
      );
      $entity = $this->create($values);

      // Set the remote data.
      $entity->entity_data = $remote_entity;

      // Set the remote ID.
      $entity->remote_id = $remote_entity->{$remote_id_key};
    }

    return $entity;
  }

  // TODO: unpack, prior to saving remotely.

}
